@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,20 +21,24 @@
 
 # This file contains functions that load and save the game state.
 
-import pickle
-import cPickle
-
-from cStringIO import StringIO
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+from future.utils import reraise
 
+import io
 import zipfile
 import re
 import threading
 import types
 import shutil
 import os
+import sys
 
 import renpy
 
+import pickle
+import renpy.compat.pickle as cPickle
+
 from json import dumps as json_dumps
 
 # Dump that chooses which pickle to use:
@@ -60,6 +64,7 @@ def loads(s):
     else:
         return pickle.loads(s)
 
+
 # This is used as a quick and dirty way of versioning savegame
 # files.
 savegame_suffix = renpy.savegame_suffix
@@ -81,10 +86,10 @@ def save_dump(roots, log):
             f.write("{0: 7d} {1} = alias {2}\n".format(0, path, o_repr_cache[ido]))
             return 0
 
-        if isinstance(o, (int, float, types.NoneType, types.ModuleType, types.ClassType)):
+        if isinstance(o, (int, float, type(None), types.ModuleType, type)):
             o_repr = repr(o)
 
-        elif isinstance(o, (str, unicode)):
+        elif isinstance(o, basestring):
             if len(o) <= 80:
                 o_repr = repr(o).encode("utf-8")
             else:
@@ -97,7 +102,7 @@ def save_dump(roots, log):
             o_repr = "<" + o.__class__.__name__ + ">"
 
         elif isinstance(o, types.MethodType):
-            o_repr = "<method {0}.{1}>".format(o.im_class.__name__, o.im_func.__name__)
+            o_repr = "<method {0}.{1}>".format(o.__self__.__class__.__name__, o.__func__.__name__)
 
         elif isinstance(o, object):
             o_repr = "<{0}>".format(type(o).__name__)
@@ -107,11 +112,11 @@ def save_dump(roots, log):
 
         o_repr_cache[ido] = o_repr
 
-        if isinstance(o, (int, float, types.NoneType, types.ModuleType, types.ClassType)):
+        if isinstance(o, (int, float, type(None), types.ModuleType, type)):
             size = 1
 
-        elif isinstance(o, (str, unicode)):
-            size = len(o) / 40 + 1
+        elif isinstance(o, bytes):
+            size = len(o) // 40 + 1
 
         elif isinstance(o, (tuple, list)):
             size = 1
@@ -121,12 +126,12 @@ def save_dump(roots, log):
 
         elif isinstance(o, dict):
             size = 2
-            for k, v in o.iteritems():
+            for k, v in o.items():
                 size += 2
                 size += visit(v, "{0}[{1!r}]".format(path, k))
 
         elif isinstance(o, types.MethodType):
-            size = 1 + visit(o.im_self, path + ".im_self")
+            size = 1 + visit(o.__self__, path + ".im_self")
 
         else:
 
@@ -150,7 +155,7 @@ def save_dump(roots, log):
 
             state = get(2, { })
             if isinstance(state, dict):
-                for k, v in state.iteritems():
+                for k, v in state.items():
                     size += 2
                     size += visit(v, path + "." + k)
             else:
@@ -174,12 +179,108 @@ def save_dump(roots, log):
 
         return size
 
-    f = file("save_dump.txt", "w")
+    f, _ = renpy.error.open_error_file("save_dump.txt", "w")
 
-    visit(roots, "roots")
-    visit(log, "log")
+    with f:
+        visit(roots, "roots")
+        visit(log, "log")
+    
+def find_bad_reduction(roots, log):
+    """
+    Finds objects that can't be reduced properly.
+    """
 
-    f.close()
+    seen = set()
+
+    def visit(o, path):
+        ido = id(o)
+
+        if ido in seen:
+            return
+
+        seen.add(ido)
+
+        if isinstance(o, (int, float, type(None), type)):
+            return
+
+        if isinstance(o, (tuple, list)):
+            for i, oo in enumerate(o):
+                rv = visit(oo, "{0}[{1!r}]".format(path, i))
+                if rv is not None:
+                    return rv
+
+        elif isinstance(o, dict):
+            for k, v in o.items():
+                rv = visit(v, "{0}[{1!r}]".format(path, k))
+                if rv is not None:
+                    return rv
+
+        elif isinstance(o, types.MethodType):
+            return visit(o.__self__, path + ".im_self")
+
+        elif isinstance(o, types.ModuleType):
+
+            return "{} = {}".format(path, repr(o)[:160])
+
+        else:
+
+            try:
+                reduction = o.__reduce_ex__(2)
+            except:
+
+                import copy
+
+                try:
+                    copy.copy(o)
+                    return None
+                except:
+                    pass
+
+                return "{} = {}".format(path, repr(o)[:160])
+
+            # Gets an element from the reduction, or o if we don't have
+            # such an element.
+            def get(idx, default):
+                if idx < len(reduction) and reduction[idx] is not None:
+                    return reduction[idx]
+                else:
+                    return default
+
+            state = get(2, { })
+            if isinstance(state, dict):
+                for k, v in state.items():
+                    rv = visit(v, path + "." + k)
+                    if rv is not None:
+                        return rv
+            else:
+                rv = visit(state, path + ".__getstate__()")
+                if rv is not None:
+                    return rv
+
+            for i, oo in enumerate(get(3, [])):
+                rv = visit(oo, "{0}[{1}]".format(path, i))
+                if rv is not None:
+                    return rv
+
+            for i in get(4, []):
+
+                if len(i) != 2:
+                    continue
+
+                k, v = i
+
+                rv = visit(v, "{0}[{1!r}]".format(path, k))
+                if rv is not None:
+                    return rv
+
+        return None
+
+    for k, v in roots.items():
+        rv = visit(v, k)
+        if rv is not None:
+            return rv
+
+    return visit(log, "renpy.game.log")
 
 ################################################################################
 # Saving
@@ -246,24 +347,21 @@ class SaveRecord(object):
             safe_rename(filename_new, filename)
             return
 
-        zf = zipfile.ZipFile(filename_new, "w", zipfile.ZIP_DEFLATED)
+        with zipfile.ZipFile(filename_new, "w", zipfile.ZIP_DEFLATED) as zf:
+            # Screenshot.
+            zf.writestr("screenshot.png", self.screenshot)
 
-        # Screenshot.
-        zf.writestr("screenshot.png", self.screenshot)
+            # Extra info.
+            zf.writestr("extra_info", self.extra_info.encode("utf-8"))
 
-        # Extra info.
-        zf.writestr("extra_info", self.extra_info.encode("utf-8"))
+            # Json
+            zf.writestr("json", self.json)
 
-        # Json
-        zf.writestr("json", self.json)
+            # Version.
+            zf.writestr("renpy_version", renpy.version)
 
-        # Version.
-        zf.writestr("renpy_version", renpy.version)
-
-        # The actual game.
-        zf.writestr("log", self.log)
-
-        zf.close()
+            # The actual game.
+            zf.writestr("log", self.log)
 
         safe_rename(filename_new, filename)
 
@@ -296,15 +394,33 @@ def save(slotname, extra_info='', mutate_flag=False):
     if renpy.config.save_dump:
         save_dump(roots, renpy.game.log)
 
-    logf = StringIO()
-    dump((roots, renpy.game.log), logf)
+    logf = io.BytesIO()
+    try:
+        dump((roots, renpy.game.log), logf)
+    except:
+
+        t, e, tb = sys.exc_info()
+
+        if mutate_flag:
+            reraise(t, e, tb)
+
+        try:
+            bad = find_bad_reduction(roots, renpy.game.log)
+        except:
+            reraise(t, e, tb)
+
+        if bad is None:
+            reraise(t, e, tb)
+
+        e.args = (e.args[0] + ' (perhaps {})'.format(bad),) + e.args[1:]
+        reraise(t, e, tb)
 
     if mutate_flag and renpy.python.mutate_flag:
         raise SaveAbort()
 
     screenshot = renpy.game.interface.get_screenshot()
 
-    json = { "_save_name" : extra_info }
+    json = { "_save_name" : extra_info, "_renpy_version" : list(renpy.version_tuple), "_version" : renpy.config.version }
 
     for i in renpy.config.save_json_callbacks:
         i(json)
@@ -318,6 +434,9 @@ def save(slotname, extra_info='', mutate_flag=False):
     clear_slot(slotname)
 
 
+# The thread used for autosave.
+autosave_thread = None
+
 # Flag that lets us know if an autosave is in progress.
 autosave_not_running = threading.Event()
 autosave_not_running.set()
@@ -326,7 +445,7 @@ autosave_not_running.set()
 autosave_counter = 0
 
 
-def autosave_thread(take_screenshot):
+def autosave_thread_function(take_screenshot):
 
     global autosave_counter
 
@@ -352,6 +471,9 @@ def autosave_thread(take_screenshot):
 
     finally:
         autosave_not_running.set()
+        if renpy.emscripten:
+            import emscripten
+            emscripten.syncfs()
 
 
 def autosave():
@@ -378,11 +500,14 @@ def autosave():
     if renpy.store.main_menu:
         return
 
+    if not renpy.store._autosave:
+        return
+
     force_autosave(True)
 
 
 # This assumes a screenshot has already been taken.
-def force_autosave(take_screenshot=False):
+def force_autosave(take_screenshot=False, block=False):
     """
     :doc: other
 
@@ -391,12 +516,25 @@ def force_autosave(take_screenshot=False):
     `take_screenshot`
         If True, a new screenshot will be taken. If False, the existing
         screenshot will be used.
+
+    `block`
+        If True, blocks until the autosave completes.
     """
 
+    global autosave_thread
+
+    if renpy.game.after_rollback or renpy.exports.in_rollback():
+        return
+
     # That is, autosave is running.
     if not autosave_not_running.isSet():
         return
 
+    # Join the autosave thread to clear resources.
+    if autosave_thread is not None:
+        autosave_thread.join()
+        autosave_thread = None
+
     # Do not save if we're in the main menu.
     if renpy.store.main_menu:
         return
@@ -405,16 +543,36 @@ def force_autosave(take_screenshot=False):
     if renpy.store._in_replay:
         return
 
-    autosave_not_running.clear()
-    t = threading.Thread(target=autosave_thread, args=(take_screenshot,))
-    t.daemon = True
-    t.start()
+    if block:
 
+        if renpy.config.auto_save_extra_info:
+            extra_info = renpy.config.auto_save_extra_info()
+        else:
+            extra_info = ""
+
+        cycle_saves("auto-", renpy.config.autosave_slots)
+
+        if take_screenshot:
+            renpy.exports.take_screenshot()
+
+        save("auto-1", extra_info=extra_info)
+
+        return
+
+    autosave_not_running.clear()
+
+    if not renpy.emscripten:
+        autosave_thread = threading.Thread(target=autosave_thread_function, args=(take_screenshot,))
+        autosave_thread.daemon = True
+        autosave_thread.start()
+    else:
+        autosave_thread_function(take_screenshot)
 
 ################################################################################
 # Loading and Slot Manipulation
 ################################################################################
 
+
 def scan_saved_game(slotname):
 
     c = get_cache(slotname)
@@ -510,6 +668,7 @@ def list_slots(regexp=None):
 
     return slots
 
+
 # A cache for newest slot info.
 newest_slot_cache = { }
 
@@ -663,6 +822,7 @@ def cycle_saves(name, count):
 # Cache
 ################################################################################
 
+
 # None is a possible value for some of the attributes.
 unknown = renpy.object.Sentinel("unknown")
 
@@ -713,6 +873,16 @@ class Cache(object):
 
         return self.screenshot
 
+    def preload(self):
+        """
+        Preloads all the save data (that won't take up a ton of memory).
+        """
+
+        self.get_mtime()
+        self.get_json()
+        self.get_screenshot()
+
+
 # A map from slotname to cache object. This is used to cache savegame scan
 # data until the slot changes.
 cache = { }
@@ -753,6 +923,16 @@ def clear_cache():
     renpy.exports.restart_interaction()
 
 
+def init():
+    """
+    Scans all the metadata from the save slot cache.
+    """
+
+    for i in list_slots():
+        if not i.startswith("_"):
+            get_cache(i).preload()
+
+
 # Save locations are places where saves are saved to or loaded from, or a
 # collection of such locations. This is the default save location.
 location = None
