@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,16 +21,20 @@
 
 # This module handles the logging of messages to a file.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import os.path
-import codecs
 import traceback
 import platform
 import time
 import tempfile
-
-import renpy
 import sys
+import io
+
+import encodings.latin_1  # @UnusedImport
+
+import renpy.config
 
 real_stdout = sys.stdout
 real_stderr = sys.stderr
@@ -93,13 +97,14 @@ class LogFile(object):
 
         try:
             base = os.environ.get("RENPY_LOG_BASE", renpy.config.logdir or renpy.config.basedir)
+
+            if base is None:
+                return False
+
             fn = os.path.join(base, self.name + ".txt")
 
             altfn = os.path.join(tempfile.gettempdir(), "renpy-" + self.name + ".txt")
 
-            if renpy.android:
-                print("Logging to", fn)
-
             if self.append:
                 mode = "a"
             else:
@@ -111,9 +116,9 @@ class LogFile(object):
             else:
 
                 try:
-                    self.file = codecs.open(fn, mode, "utf-8")
+                    self.file = io.open(fn, mode, encoding="utf-8")
                 except:
-                    self.file = codecs.open(altfn, mode, "utf-8")
+                    self.file = io.open(altfn, mode, encoding="utf-8")
 
             if self.append:
                 self.write('')
@@ -144,14 +149,16 @@ class LogFile(object):
         if self.open():
 
             if not self.raw_write:
-                s = s % args
+                try:
+                    s = s % args
+                except:
+                    s = repr((s,) + args)
+
                 s += "\n"
 
-            if not isinstance(s, unicode):
+            if not isinstance(s, str):
                 s = s.decode("latin-1")
 
-            s = s.replace("\n", "\r\n")
-
             self.file.write(s)
 
             if self.flush:
@@ -180,6 +187,32 @@ def open(name, append=False, developer=False, flush=False):  # @ReservedAssignme
 
     return rv
 
+################################################################################
+# Timed event log.
+
+
+class TimeLog(list):
+    """
+    This represents a log that is limited to the last `duration` seconds.
+    """
+
+    def __init__(self, duration):
+        self.duration = duration
+
+    def append(self, v):
+        now = time.time()
+
+        list.append(self, (now, v))
+        self.prune(now)
+
+    def prune(self, now=None):
+
+        if now is None:
+            now = time.time()
+
+        while self[0][0] < (now - self.duration):
+            self.pop(0)
+
 
 ################################################################################
 # Stdout / Stderr Redirection
@@ -188,9 +221,13 @@ class StdioRedirector(object):
 
     def __init__(self):
         self.buffer = ''
-        self.log = open("log", developer=False, append=False)
+        self.log = open("log", developer=False, append=False, flush=True)
 
     def write(self, s):
+
+        if not isinstance(s, str):
+            s = str(s, "utf-8", "replace")
+
         self.real_file.write(s)
         self.real_file.flush()
 
@@ -213,7 +250,6 @@ class StdioRedirector(object):
                 try:
                     i(l)
                 except:
-                    # traceback.print_exc(None, real_stderr)
                     pass
 
         self.buffer = lines[-1]
@@ -230,24 +266,28 @@ class StdioRedirector(object):
         pass
 
 
-class StdoutRedirector(StdioRedirector):
-    real_file = real_stdout
+if not "RENPY_NO_REDIRECT_STDIO" in os.environ:
 
-    def get_callbacks(self):
-        return renpy.config.stdout_callbacks
+    class StdoutRedirector(StdioRedirector):
+        real_file = real_stdout
 
+        def get_callbacks(self):
+            return renpy.config.stdout_callbacks
 
-sys.stdout = sys_stdout = StdoutRedirector()
+    sys.stdout = sys_stdout = StdoutRedirector()
 
+    class StderrRedirector(StdioRedirector):
+        real_file = real_stderr
 
-class StderrRedirector(StdioRedirector):
-    real_file = real_stderr
+        def get_callbacks(self):
+            return renpy.config.stderr_callbacks
 
-    def get_callbacks(self):
-        return renpy.config.stderr_callbacks
+    sys.stderr = sys_stderr = StderrRedirector()
 
+else:
 
-sys.stderr = sys_stderr = StderrRedirector()
+    sys_stdout = sys.stdout
+    sys_stderr = sys.stderr
 
 
 def post_init():
