@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,10 +23,11 @@
 # contained within the script file. It also handles rolling back the
 # game state to some time in the past.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 # Import the python ast module, not ours.
-ast = __import__("ast", { })
+import ast
 
 # Import the future module itself.
 import __future__
@@ -37,9 +38,38 @@ import weakref
 import re
 import sys
 import time
+import io
+import types
+import copyreg
+import functools
 
 import renpy.audio
 
+# A set of flags that indicate dict should run in future-compatible mode.
+FUTURE_FLAGS = (__future__.CO_FUTURE_DIVISION | __future__.CO_FUTURE_WITH_STATEMENT)
+
+##############################################################################
+# Monkeypatch copy_reg to work around a change in the class that RevertableSet
+# is based on.
+
+
+def _reconstructor(cls, base, state):
+    if (cls is RevertableSet) and (base is object):
+        base = set
+        state = [ ]
+
+    if base is object:
+        obj = object.__new__(cls)
+    else:
+        obj = base.__new__(cls, state)
+        if base.__init__ != object.__init__:
+            base.__init__(obj, state)
+
+    return obj
+
+
+copyreg._reconstructor = _reconstructor
+
 ##############################################################################
 # Code that implements the store.
 
@@ -50,7 +80,10 @@ import renpy.audio
 class StoreDeleted(object):
 
     def __reduce__(self):
-        return "deleted"
+        if PY2:
+            return b"deleted"
+        else:
+            return "deleted"
 
 
 deleted = StoreDeleted()
@@ -164,7 +197,7 @@ def begin_stores():
     Calls .begin on every store dict.
     """
 
-    for sd in store_dicts.itervalues():
+    for sd in store_dicts.values():
         sd.begin()
 
 
@@ -201,25 +234,27 @@ def create_store(name):
     d = store_dicts.setdefault(name, StoreDict())
     d.reset()
 
+    pyname = pystr(name)
+
     # Set the name.
-    d["__name__"] = name
-    d["__package__"] = name
+    d["__name__"] = pyname
+    d["__package__"] = pyname
 
     # Set up the default contents of the store.
     eval("1", d)
 
-    for k, v in renpy.minstore.__dict__.iteritems():
+    for k, v in renpy.minstore.__dict__.items():
         if k not in d:
             d[k] = v
 
     # Create or reuse the corresponding module.
     if name in store_modules:
-        sys.modules[name] = store_modules[name]
+        sys.modules[pyname] = store_modules[name]
     else:
-        store_modules[name] = sys.modules[name] = StoreModule(d)
+        store_modules[name] = sys.modules[pyname] = StoreModule(d)
 
     if parent:
-        store_dicts[parent][var] = sys.modules[name]
+        store_dicts[parent][var] = sys.modules[pyname]
 
 
 class StoreBackup():
@@ -277,7 +312,7 @@ def make_clean_stores():
 
     global clean_store_backup
 
-    for _k, v in store_dicts.iteritems():
+    for _k, v in store_dicts.items():
 
         v.ever_been_changed.clear()
         v.begin()
@@ -327,18 +362,16 @@ class NoRollback(object):
 
     pass
 
-
 # parents = [ ]
 
+
 def reached(obj, reachable, wait):
     """
     @param obj: The object that was reached.
-    @param path: The path from the store via which it was reached.
 
     `reachable`
         A map from id(obj) to int. The int is 1 if the object was reached
         normally, and 0 if it was reached, but inherits from NoRollback.
-
     """
 
     if wait:
@@ -349,7 +382,7 @@ def reached(obj, reachable, wait):
     if idobj in reachable:
         return
 
-    if isinstance(obj, (NoRollback, real_file)):  # @UndefinedVariable
+    if isinstance(obj, (NoRollback, io.IOBase)): # @UndefinedVariable
         reachable[idobj] = 0
         return
 
@@ -364,7 +397,7 @@ def reached(obj, reachable, wait):
 
     try:
         # Treat as fields, indexed by strings.
-        for v in vars(obj).itervalues():
+        for v in vars(obj).values():
             reached(v, reachable, wait)
     except:
         pass
@@ -379,7 +412,7 @@ def reached(obj, reachable, wait):
 
     try:
         # Treat as dict.
-        for v in obj.itervalues():
+        for v in obj.values():
             reached(v, reachable, wait)
     except:
         pass
@@ -397,33 +430,40 @@ def reached_vars(store, reachable, wait):
     the path by which the object was reached.
     """
 
-    for v in store.itervalues():
+    for v in store.values():
         reached(v, reachable, wait)
 
     for c in renpy.game.contexts:
         reached(c.info, reachable, wait)
         reached(c.music, reachable, wait)
         for d in c.dynamic_stack:
-            for v in d.itervalues():
+            for v in d.values():
                 reached(v, reachable, wait)
 
-
 # Code that replaces literals will calls to magic constructors.
 
+
+def b(s):
+    if PY2:
+        return s.encode("utf-8")
+    else:
+        return s
+
+
 class WrapNode(ast.NodeTransformer):
 
     def visit_ClassDef(self, n):
         n = self.generic_visit(n)
 
         if not n.bases:
-            n.bases.append(ast.Name(id="object", ctx=ast.Load()))
+            n.bases.append(ast.Name(id=b("object"), ctx=ast.Load()))
 
         return n
 
     def visit_SetComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__set__",
+                id=b("__renpy__set__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -435,7 +475,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__set__",
+                id=b("__renpy__set__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -446,7 +486,7 @@ class WrapNode(ast.NodeTransformer):
     def visit_ListComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__list__",
+                id=b("__renpy__list__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -460,7 +500,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__list__",
+                id=b("__renpy__list__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -471,7 +511,7 @@ class WrapNode(ast.NodeTransformer):
     def visit_DictComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__dict__",
+                id=b("__renpy__dict__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -483,7 +523,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__dict__",
+                id=b("__renpy__dict__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -514,7 +554,7 @@ _execute_python_hide()
     tree.body = hide.body
 
 
-unicode_re = re.compile(ur'[\u0080-\uffff]')
+unicode_re = re.compile(r'[\u0080-\uffff]')
 
 
 def unicode_sub(m):
@@ -551,16 +591,22 @@ def escape_unicode(s):
 
 
 # Flags used by py_compile.
-old_compile_flags = ( __future__.nested_scopes.compiler_flag
+old_compile_flags = (__future__.nested_scopes.compiler_flag
                       | __future__.with_statement.compiler_flag
                       )
 
-new_compile_flags = (  old_compile_flags
-                       | __future__.absolute_import.compiler_flag
-                       | __future__.print_function.compiler_flag
-                       | __future__.unicode_literals.compiler_flag
-                       )
+new_compile_flags = (old_compile_flags
+                      | __future__.absolute_import.compiler_flag
+                      | __future__.print_function.compiler_flag
+                      | __future__.unicode_literals.compiler_flag
+                      )
 
+py3_compile_flags = (old_compile_flags |
+                      __future__.division.compiler_flag)
+
+# The set of files that should be compiled under Python 2 with Python 3
+# semantics.
+py3_files = set()
 
 # A cache for the results of py_compile.
 py_compile_cache = { }
@@ -622,7 +668,7 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
         lineno = source.linenumber
 
     if cache:
-        key = (lineno, filename, unicode(source), mode, renpy.script.MAGIC)
+        key = (lineno, filename, str(source), mode, renpy.script.MAGIC)
 
         rv = py_compile_cache.get(key, None)
         if rv is not None:
@@ -630,7 +676,7 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
 
         rv = old_py_compile_cache.get(key, None)
         if rv is not None:
-            old_py_compile_cache[key] = rv
+            py_compile_cache[key] = rv
             return rv
 
         bytecode = renpy.game.script.bytecode_oldcache.get(key, None)
@@ -641,10 +687,13 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
             py_compile_cache[key] = rv
             return rv
 
-    source = unicode(source)
+    source = str(source)
     source = source.replace("\r", "")
     source = escape_unicode(source)
 
+    if mode == "eval":
+        source = source.replace("\n", "\\\n").replace("\\\\\n", "\\\n")
+
     try:
         line_offset = lineno - 1
 
@@ -653,13 +702,20 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
         else:
             py_mode = mode
 
-        try:
-            flags = new_compile_flags
-            tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
-        except:
-            flags = old_compile_flags
+        if filename in py3_files:
+
+            flags = py3_compile_flags
             tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
 
+        else:
+
+            try:
+                flags = new_compile_flags
+                tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
+            except:
+                flags = old_compile_flags
+                tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
+
         tree = wrap_node.visit(tree)
 
         if mode == "hide":
@@ -705,10 +761,10 @@ def py_compile_eval_bytecode(source, **kwargs):
     code = py_compile(source, 'eval', cache=False, **kwargs)
     return marshal.dumps(code)
 
-
 # Classes that are exported in place of the normal list, dict, and
 # object.
 
+
 # This is set to True whenever a mutation occurs. The save code uses
 # this to check to see if a background-save is valid.
 mutate_flag = True
@@ -716,14 +772,15 @@ mutate_flag = True
 
 def mutator(method):
 
+    @functools.wraps(method, ("__name__", "__doc__"), ())
     def do_mutation(self, *args, **kwargs):
 
         global mutate_flag
 
-        mutated = renpy.game.log.mutated  # @UndefinedVariable
+        mutated = renpy.game.log.mutated # @UndefinedVariable
 
         if id(self) not in mutated:
-            mutated[id(self)] = ( weakref.ref(self), self._clean())
+            mutated[id(self)] = (weakref.ref(self), self._clean())
             mutate_flag = True
 
         return method(self, *args, **kwargs)
@@ -788,10 +845,10 @@ class CompressedList(object):
             old_end += 1
 
         # Now that we have this, we can put together the object.
-        self.pre = list.__getslice__(old, 0, old_start)
+        self.pre = list.__getitem__(old, slice(0, old_start))
         self.start = new_start
         self.end = new_end
-        self.post = list.__getslice__(old, old_end, len_old)
+        self.post = list.__getitem__(old, slice(old_end, len_old))
 
     def decompress(self, new):
         return self.pre + new[self.start:self.end] + self.post
@@ -815,8 +872,11 @@ class RevertableList(list):
         list.__init__(self, *args)
 
     __delitem__ = mutator(list.__delitem__)
-    __delslice__ = mutator(list.__delslice__)
+    if PY2:
+        __delslice__ = mutator(list.__delslice__)
     __setitem__ = mutator(list.__setitem__)
+    if PY2:
+        __setslice__ = mutator(list.__setslice__)
     __iadd__ = mutator(list.__iadd__)
     __imul__ = mutator(list.__imul__)
     append = mutator(list.append)
@@ -827,7 +887,8 @@ class RevertableList(list):
     reverse = mutator(list.reverse)
     sort = mutator(list.sort)
 
-    def wrapper(method):  # E0213 @NoSelf
+    def wrapper(method): # E0213 @NoSelf
+
         def newmethod(*args, **kwargs):
             l = method(*args, **kwargs)
             return RevertableList(l)
@@ -835,7 +896,16 @@ class RevertableList(list):
         return newmethod
 
     __add__ = wrapper(list.__add__)
-    __getslice__ = wrapper(list.__getslice__)
+    if PY2:
+        __getslice__ = wrapper(list.__getslice__)
+
+    def __getitem__(self, index):
+        rv = list.__getitem__(self, index)
+
+        if isinstance(index, slice):
+            return RevertableList(rv)
+        else:
+            return rv
 
     def __mul__(self, other):
         if not isinstance(other, int):
@@ -845,6 +915,12 @@ class RevertableList(list):
 
     __rmul__ = __mul__
 
+    def copy(self):
+        return self[:]
+
+    def clear(self):
+        self[:] = []
+
     def _clean(self):
         """
         Gets a clean copy of this object before any mutation occurs.
@@ -909,17 +985,31 @@ class RevertableDict(dict):
     setdefault = mutator(dict.setdefault)
     update = mutator(dict.update)
 
-    def list_wrapper(method):  # E0213 @NoSelf
-        def newmethod(*args, **kwargs):
-            return RevertableList(method(*args, **kwargs))  # E1102
+    if PY2:
 
-        return newmethod
+        def keys(self):
+            rv = dict.keys(self)
 
-    keys = list_wrapper(dict.keys)
-    values = list_wrapper(dict.values)
-    items = list_wrapper(dict.items)
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
 
-    del list_wrapper
+            return rv
+
+        def values(self):
+            rv = dict.values(self)
+
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
+
+            return rv
+
+        def items(self):
+            rv = dict.items(self)
+
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
+
+            return rv
 
     def copy(self):
         rv = RevertableDict()
@@ -927,7 +1017,7 @@ class RevertableDict(dict):
         return rv
 
     def _clean(self):
-        return self.items()
+        return list(self.items())
 
     def _compress(self, clean):
         return clean
@@ -948,7 +1038,7 @@ class RevertableSet(set):
             self.update(state)
 
     def __getstate__(self):
-        rv = ({ i : True for i in self}, )
+        rv = ({ i : True for i in self},)
         return rv
 
     # Required to ensure that getstate and setstate are called.
@@ -978,7 +1068,8 @@ class RevertableSet(set):
     union_update = mutator(set.update)
     update = mutator(set.update)
 
-    def wrapper(method):  # @NoSelf
+    def wrapper(method): # @NoSelf
+
         def newmethod(*args, **kwargs):
             rv = method(*args, **kwargs)
             if isinstance(rv, (set, frozenset)):
@@ -1046,9 +1137,54 @@ class RevertableObject(object):
         self.__dict__.update(compressed)
 
 
-# An object that handles deterministic randomness, or something.
+class RollbackRandom(random.Random):
+    """
+    This is used for Random objects returned by renpy.random.Random.
+    """
+
+    def __init__(self):
+        log = renpy.game.log
+
+        if log is not None:
+            log.mutated[id(self)] = None
+
+        super(RollbackRandom, self).__init__()
+
+    def _clean(self):
+        return self.getstate()
+
+    def _compress(self, clean):
+        return clean
+
+    def _rollback(self, compressed):
+        super(RollbackRandom, self).setstate(compressed)
+
+    setstate = mutator(random.Random.setstate)
+
+    if PY2:
+        jumpahead = mutator(random.Random.jumpahead)
+
+    getrandbits = mutator(random.Random.getrandbits)
+    seed = mutator(random.Random.seed)
+    random = mutator(random.Random.random)
+
+    def Random(self, seed=None):
+        """
+        Returns a new RNG object separate from the main one.
+        """
+
+        if seed is None:
+            seed = self.random()
+
+        new = RollbackRandom()
+        new.seed(seed)
+        return new
+
 
 class DetRandom(random.Random):
+    """
+    This is renpy.random.
+    """
 
     def __init__(self):
         super(DetRandom, self).__init__()
@@ -1084,14 +1220,17 @@ class DetRandom(random.Random):
         Resets the RNG, removing all of the pushbacked numbers.
         """
 
-        self.stack = [ ]
+        del self.stack[:]
 
     def Random(self, seed=None):
         """
         Returns a new RNG object separate from the main one.
         """
 
-        new = DetRandom()
+        if seed is None:
+            seed = self.random()
+
+        new = RollbackRandom()
         new.seed(seed)
         return new
 
@@ -1211,15 +1350,15 @@ class Rollback(renpy.object.Object):
 
         # Add objects reachable from the stores. (Objects that might be
         # unreachable at the moment.)
-        for changes in self.stores.itervalues():
-            for _k, v in changes.iteritems():
+        for changes in self.stores.values():
+            for _k, v in changes.items():
                 if v is not deleted:
                     reached(v, reachable, wait)
 
         # Add in objects reachable through the context.
         reached(self.context.info, reachable, wait)
         for d in self.context.dynamic_stack:
-            for v in d.itervalues():
+            for v in d.values():
                 reached(v, reachable, wait)
 
         # Add in objects reachable through displayables.
@@ -1228,16 +1367,26 @@ class Rollback(renpy.object.Object):
         # Purge object update information for unreachable objects.
         new_objects = [ ]
 
-        for o, rb in self.objects:
-            if reachable.get(id(o), 0):
-                new_objects.append((o, rb))
-                reached(rb, reachable, wait)
-            else:
-                if renpy.config.debug:
-                    print("Removing unreachable:", o, file=renpy.log.real_stdout)
-                    pass
+        objects_changed = True
+        seen = set()
 
-        self.objects = new_objects
+        while objects_changed:
+
+            objects_changed = False
+
+            for o, rb in self.objects:
+
+                id_o = id(o)
+
+                if (id_o not in seen) and reachable.get(id_o, 0):
+                    seen.add(id_o)
+                    objects_changed = True
+
+                    new_objects.append((o, rb))
+                    reached(rb, reachable, wait)
+
+        del self.objects[:]
+        self.objects.extend(new_objects)
 
         return True
 
@@ -1251,19 +1400,19 @@ class Rollback(renpy.object.Object):
             if roll is not None:
                 obj._rollback(roll)
 
-        for name, changes in self.stores.iteritems():
+        for name, changes in self.stores.items():
             store = store_dicts.get(name, None)
             if store is None:
                 continue
 
-            for name, value in changes.iteritems():
+            for name, value in changes.items():
                 if value is deleted:
                     if name in store:
                         del store[name]
                 else:
                     store[name] = value
 
-        for name, changes in self.delta_ebc.iteritems():
+        for name, changes in self.delta_ebc.items():
 
             store = store_dicts.get(name, None)
             if store is None:
@@ -1308,6 +1457,7 @@ class RollbackLog(renpy.object.Object):
 
     nosave = [ 'old_store', 'mutated', 'identifier_cache' ]
     identifier_cache = None
+    force_checkpoint = False
 
     def __init__(self):
 
@@ -1337,6 +1487,10 @@ class RollbackLog(renpy.object.Object):
         # reset?
         self.did_interaction = True
 
+        # Should we force a checkpoint before completing the current
+        # statement.
+        self.force_checkpoint = False
+
     def after_setstate(self):
         self.mutated = { }
         self.rolled_forward = False
@@ -1401,8 +1555,8 @@ class RollbackLog(renpy.object.Object):
             begin_stores()
 
         # If the log is too long, prune it.
-        if len(self.log) > renpy.config.rollback_length:
-            self.log = self.log[-renpy.config.rollback_length:]
+        while len(self.log) > renpy.config.rollback_length:
+            self.log.pop(0)
 
         # check for the end of fixed rollback
         if self.log and self.log[-1] == self.current:
@@ -1420,7 +1574,7 @@ class RollbackLog(renpy.object.Object):
 
         self.log.append(self.current)
 
-        self.mutated = { }
+        self.mutated.clear()
 
         # Flag a mutation as having happened. This is used by the
         # save code.
@@ -1450,20 +1604,24 @@ class RollbackLog(renpy.object.Object):
             Should be true if called from begin().
         """
 
+        if self.force_checkpoint:
+            self.checkpoint(hard=False)
+            self.force_checkpoint = False
+
         # Update self.current.stores with the changes from each store.
         # Also updates .ever_been_changed.
-        for name, sd in store_dicts.iteritems():
+        for name, sd in store_dicts.items():
             self.current.stores[name], self.current.delta_ebc[name] = sd.get_changes(begin)
 
         # Update the list of mutated objects and what we need to do to
         # restore them.
 
-        for _i in xrange(4):
+        for _i in range(4):
 
-            self.current.objects = [ ]
+            del self.current.objects[:]
 
             try:
-                for _k, v in self.mutated.iteritems():
+                for _k, v in self.mutated.items():
 
                     if v is None:
                         continue
@@ -1494,7 +1652,7 @@ class RollbackLog(renpy.object.Object):
 
         rv = { }
 
-        for store_name, sd in store_dicts.iteritems():
+        for store_name, sd in store_dicts.items():
             for name in sd.ever_been_changed:
                 if name in sd:
                     rv[store_name + "." + name] = sd[name]
@@ -1582,7 +1740,7 @@ class RollbackLog(renpy.object.Object):
                 self.forward.pop(0)
             else:
                 self.current.forward = data
-                self.forward = [ ]
+                del self.forward[:]
 
         elif data is not None:
             if self.forward:
@@ -1596,7 +1754,7 @@ class RollbackLog(renpy.object.Object):
                     ):
                     self.forward.pop(0)
                 else:
-                    self.forward = [ ]
+                    del self.forward[:]
 
             # Log the data in case we roll back again.
             self.current.forward = data
@@ -1619,7 +1777,7 @@ class RollbackLog(renpy.object.Object):
         renpy.game.context().force_checkpoint = True
 
         if purge:
-            self.log = [ ]
+            del self.log[:]
 
     def retain_after_load(self):
         """
@@ -1645,6 +1803,29 @@ class RollbackLog(renpy.object.Object):
 
         return self.rollback_limit > 0
 
+    def load_failed(self):
+        """
+        This is called to try to recover when rollback fails.
+        """
+
+        lfl = renpy.config.load_failed_label
+        if callable(lfl):
+            lfl = lfl()
+
+        if not lfl:
+            raise Exception("Couldn't find a place to stop rolling back. Perhaps the script changed in an incompatible way?")
+
+        rb = self.log.pop()
+        rb.rollback()
+
+        while renpy.exports.call_stack_depth():
+            renpy.exports.pop_call()
+
+        renpy.game.contexts[0].force_checkpoint = True
+        renpy.game.contexts[0].goto_label(lfl)
+
+        raise renpy.game.RestartTopContext()
+
     def rollback(self, checkpoints, force=False, label=None, greedy=True, on_load=False, abnormal=True, current_label=None):
         """
         This rolls the system back to the first valid rollback point
@@ -1671,7 +1852,7 @@ class RollbackLog(renpy.object.Object):
             load. Used to implement .retain_after_load()
 
         `abnormal`
-            If true, treats this as an abnormal event, suppresisng rollback
+            If true, treats this as an abnormal event, suppressing transitions
             and so on.
 
         `current_label`
@@ -1708,15 +1889,16 @@ class RollbackLog(renpy.object.Object):
                     break
 
         else:
-            if force:
-                raise Exception("Couldn't find a place to stop rolling back. Perhaps the script changed in an incompatible way?")
-
             # Otherwise, just give up.
 
-            print("Can't find a place to rollback to. Not rolling back.")
-
             revlog.reverse()
-            self.log = self.log + revlog
+            self.log.extend(revlog)
+
+            if force:
+                self.load_failed()
+            else:
+                print("Can't find a place to rollback to. Not rolling back.")
+
             return
 
         force_checkpoint = False
@@ -1783,7 +1965,7 @@ class RollbackLog(renpy.object.Object):
         # If necessary, reset the RNG.
         if force:
             rng.reset()
-            self.forward = [ ]
+            del self.forward[:]
 
         # Flag that we're in the transition immediately after a rollback.
         renpy.game.after_rollback = abnormal
@@ -1851,7 +2033,7 @@ class RollbackLog(renpy.object.Object):
         """
 
         # Fix up old screens.
-        renpy.display.screen.before_restart()  # @UndefinedVariable
+        renpy.display.screen.before_restart() # @UndefinedVariable
 
         # Set us up as the game log.
         renpy.game.log = self
@@ -1859,7 +2041,7 @@ class RollbackLog(renpy.object.Object):
         clean_stores()
         renpy.translation.init_translation()
 
-        for name, value in roots.iteritems():
+        for name, value in roots.items():
 
             if "." in name:
                 store_name, name = name.rsplit(".", 1)
@@ -1915,18 +2097,18 @@ class RollbackLog(renpy.object.Object):
         return self.identifier_cache.get(identifier, None)
 
 
-def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"):  # @ReservedAssignment
+def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"): # @ReservedAssignment
 
     if hide:
-        locals = { }  # @ReservedAssignment
+        locals = { } # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts[store]  # @ReservedAssignment
+        globals = store_dicts[store] # @ReservedAssignment
 
     if locals is None:
-        locals = globals  # @ReservedAssignment
+        locals = globals # @ReservedAssignment
 
-    exec bytecode in globals, locals
+    exec(bytecode, globals, locals)
 
 
 def py_exec(source, hide=False, store=None):
@@ -1935,25 +2117,25 @@ def py_exec(source, hide=False, store=None):
         store = store_dicts["store"]
 
     if hide:
-        locals = { }  # @ReservedAssignment
+        locals = { } # @ReservedAssignment
     else:
-        locals = store  # @ReservedAssignment
+        locals = store # @ReservedAssignment
 
-    exec py_compile(source, 'exec') in store, locals
+    exec(py_compile(source, 'exec'), store, locals)
 
 
-def py_eval_bytecode(bytecode, globals=None, locals=None):  # @ReservedAssignment
+def py_eval_bytecode(bytecode, globals=None, locals=None): # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts["store"]  # @ReservedAssignment
+        globals = store_dicts["store"] # @ReservedAssignment
 
     if locals is None:
-        locals = globals  # @ReservedAssignment
+        locals = globals # @ReservedAssignment
 
     return eval(bytecode, globals, locals)
 
 
-def py_eval(code, globals=None, locals=None):  # @ReservedAssignment
+def py_eval(code, globals=None, locals=None): # @ReservedAssignment
     if isinstance(code, basestring):
         code = py_compile(code, 'eval')
 
@@ -1983,7 +2165,7 @@ def raise_at_location(e, loc):
     code = compile(node, filename, 'exec')
 
     # PY3 - need to change to exec().
-    exec code in { "e" : e }
+    exec(code, { "e" : e })
 
 
 # This was used to proxy accesses to the store. Now it's kept around to deal
@@ -1991,29 +2173,32 @@ def raise_at_location(e, loc):
 class StoreProxy(object):
 
     def __getattr__(self, k):
-        return getattr(renpy.store, k)  # @UndefinedVariable
+        return getattr(renpy.store, k) # @UndefinedVariable
 
     def __setattr__(self, k, v):
-        setattr(renpy.store, k, v)  # @UndefinedVariable
+        setattr(renpy.store, k, v) # @UndefinedVariable
 
     def __delattr__(self, k):
-        delattr(renpy.store, k)  # @UndefinedVariable
+        delattr(renpy.store, k) # @UndefinedVariable
 
 
-# Code for pickling bound methods.
-def method_pickle(method):
-    name = method.im_func.__name__
+if PY2:
 
-    obj = method.im_self
+    # Code for pickling bound methods.
+    def method_pickle(method):
+        name = method.im_func.__name__
 
-    if obj is None:
-        obj = method.im_class
+        obj = method.im_self
 
-    return method_unpickle, (obj, name)
+        if obj is None:
+            obj = method.im_class
 
+        return method_unpickle, (obj, name)
 
-def method_unpickle(obj, name):
-    return getattr(obj, name)
+    def method_unpickle(obj, name):
+        return getattr(obj, name)
+
+    copyreg.pickle(types.MethodType, method_pickle)
 
 # Code for pickling modules.
 
@@ -2029,8 +2214,4 @@ def module_unpickle(name):
     return __import__(name)
 
 
-import copy_reg
-import types
-
-copy_reg.pickle(types.MethodType, method_pickle)
-copy_reg.pickle(types.ModuleType, module_pickle)
+copyreg.pickle(types.ModuleType, module_pickle)
